% This is the main run file for replicating the application in Chan and Qi (2025).
% This code estimates 4 BMARs:
% 1. homoskedastic BMAR (BMAR);
% 2. BMAR with the common SV (BMAR-CSV);
% 3. BMAR with t errors (BMAR-t);
% 4. BMAR with an outlier component  (BMAR-O)
% It also computes the marginal likelihood of BMAR.

% This code is free to use for academic purposes only, provided that the 
% paper is cited as:
%
% Chan, J.C.C. and Qi, Y. (2025). Large Bayesian Matrix Autoregressions, 
% Journal of Econometrics, forthcoming
%
% This code comes without technical support of any kind.  It is expected to
% reproduce the results reported in the paper. Under no circumstances will
% the authors be held responsible for any use (or misuse) of this code in
% any way.

clear; clc;
addpath('./utility') 
nsim = 10000; burnin = 500;
T0 = 8; 
p = 2;
dataset = 2;
cp_ml = 1;   % 1: compute marginal likelihood - only available for BMAR
R = 5000;    % # of importance sampling draws for computing the marginal likelihood 
model = 'BMAR-t'; % BMAR, BMAR-CSV, BMAR-t, BMAR-O
if dataset == 2 
    filename = 'Data_state_level_6q.xlsx';
    [data,regionID] = CleanData6q(filename);        
    MuY = mean(data,3);  
    data = data - repmat(MuY,[1,1,size(data,3)]);  % demean the data (no intercepts for baseline model)
end
Y0 = data(:,:,1:T0);  % use the first T0 obs as initial values
Y = data(:,:,T0+1:end); 
[n,k,T] = size(Y);

    % construct XA, XB, YA, YB
XA = zeros(n*p,T*k*p); YA = zeros(n,T*k);
XB = zeros(k*p,T*n*p); YB = zeros(k,T*n);
for t = (T0+1):(T+T0)
    Xtcell = num2cell(data(:,:,t-1:-1:t-p),[1,2]);
    Xt = blkdiag(Xtcell{:});
    XA(:,(t-T0-1)*k*p+1:(t-T0)*k*p) = Xt;
    YA(:,(t-T0-1)*k+1:(t-T0)*k) = Y(:,:,t-T0);    
    XB(:,(t-T0-1)*n*p+1:(t-T0)*n*p) = Xt';
    YB(:,(t-T0-1)*n+1:(t-T0)*n) = Y(:,:,t-T0)';
end

    % priors
kappa = [.02,1]'; % kappaA, kappaB;
[Hyper.A0,Hyper.VA,Hyper.B0,Hyper.VB,Hyper.nur,Hyper.Sr,Hyper.nuc,Hyper.Sc,U_hat] = prior_MAR_NCP(p,kappa(1),kappa(2),Y0,Y);
CA = Hyper.VA/kappa(1); % the constant part of VA
CB = Hyper.VB/kappa(2);
Hyper.iVA = sparse(1:n*p,1:n*p,1./Hyper.VA);
Hyper.iVB = sparse(1:k*p,1:k*p,1./Hyper.VB);
Hyper.cA0 = [5,5]; 
Hyper.cB0 = [5,5]; 
Hyper.nuh = 5; Hyper.Sh = 0.1*(Hyper.nuh-1);
Hyper.phi0 = .95; Hyper.Vphi = .25^2;
Hyper.nuub = 50; % upperbound on nu (BMAR-t)
Hyper.p0a = 10/4; Hyper.p0b = (1-1/16)*40; % hyperparameters for p0 (BMAR-O)

    % storage initialization
store_A = zeros(nsim,n*p,n);
store_B = zeros(nsim,k*p,k);
store_Sigc = zeros(nsim,k,k);
store_Sigr = zeros(nsim,n,n);
store_omega = zeros(nsim,T);
store_kappa = zeros(nsim,2);
if strcmp(model,'BMAR') 
    store_theta = zeros(nsim,2); % (kappa_A,kappa_B)    
elseif strcmp(model,'BMAR-CSV') 
    store_theta = zeros(nsim,2); % (phi,sigh2)
elseif strcmp(model,'BMAR-t') 
    store_theta = zeros(nsim,1); % nu
elseif strcmp(model,'BMAR-O')
    store_theta = zeros(nsim,1); % po
end

start_time = clock;
    % MCMC starts here
disp(['Starting MCMC for ' model '...']);
randn('seed',sum(clock*100)); rand('seed',sum(clock*1000));    
    % initialize the Markov chain
B = repmat(eye(k),p,1);
Sigc = eye(k); iSigc = Sigc\speye(k);
if strcmp(model,'BMAR') 
    omega = ones(T,1);
elseif strcmp(model,'BMAR-CSV') 
    phi = .97;
    sigh2 = .05;
    h = genSV(0,phi,sigh2,T);
    h = h-mean(h);
    omega = exp(h);   
    count_h = 0; is_accept = 0;
elseif strcmp(model,'BMAR-t') 
    nu = 2;
    omega = 1./gamrnd(nu/2,2/nu,T,1);
    count_nu = 0; is_accept = 0;
elseif strcmp(model,'BMAR-O')
    po = 1/16;     % outlier probability
    o = ones(T,1);
    omega = o.^2;
    ngrid = 31;    % # of grid points in the grid for o
    o_grid = [1;linspace(2,20,ngrid)'];  % grid for o
end
iOmega = sparse(1:T,1:T,1./omega);    

% construct a few things
    % for imposing the constraints M*vec(B) = b0
M = sparse(1:p,1:k:k*p,ones(p,1),p,k^2*p);
b0 = ones(p,1);

for isim = 1:nsim + burnin
        % sample (A, Sigr)    
    KA = Hyper.iVA + XA*kron(iOmega,B*iSigc*B')*XA';
    CKA = chol(KA,'lower');
    Ahat = CKA'\(CKA\(Hyper.iVA*Hyper.A0 + XA*kron(iOmega,B*iSigc)*YA'));    
    Srhat = Hyper.Sr + Hyper.A0'*Hyper.iVA*Hyper.A0 + YA*kron(iOmega,iSigc)*YA' - Ahat'*KA*Ahat;    
    Srhat = (Srhat+Srhat')/2;
    Sigr = iwishrnd(Srhat, Hyper.nur + T*k);
    CSigr = chol(Sigr,"lower");
    A = Ahat + (CKA'\randn(n*p,n))*CSigr';
    iSigr = Sigr\speye(n);    

        % sample (B, Sigc)
    KB = Hyper.iVB + XB*kron(iOmega,A*iSigr*A')*XB';
    CKB = chol(KB,'lower');
    Bhat = CKB'\(CKB\(Hyper.iVB*Hyper.B0 + XB*kron(iOmega,A*iSigr)*YB'));
    Schat = Hyper.Sc + Hyper.B0'*Hyper.iVB*Hyper.B0 + YB*kron(iOmega,iSigr)*YB' - Bhat'*KB*Bhat;  
    Schat = (Schat+Schat')/2;
    Sigc = nobile_iwishart(k,Hyper.nuc + T*n, Schat);    
    iSigc = Sigc\speye(k);
    CSigc = chol(Sigc,"lower");
    CiSigc = chol(iSigc,"lower");
    B = Bhat + (CKB'\randn(k*p,k))*CSigc';    
    vecB = B(:);    
    C = kron(CiSigc,CKB); % C = kron(CSigc\speye(k),CKB);
    tmpU = C'\(C\M');
    tmpV = (M*tmpU)\tmpU';
    vecB = vecB - tmpV'*(M*vecB-b0);
    B = reshape(vecB,k*p,k); 
    
        % sample kappaA and kappaB
    QA = diag((A-Hyper.A0)*(Sigr\(A-Hyper.A0)'));
    tmpA = sum(QA./CA);
    QB = diag((B-Hyper.B0)*(Sigc\(B-Hyper.B0)'));
    tmpB = sum(QB./CB);    
    kappa(1) = gigrnd(Hyper.cA0(1)-n^2*p/2,2*Hyper.cA0(2),tmpA,1);
    kappa(2) = gigrnd(Hyper.cB0(1)-k^2*p/2,2*Hyper.cB0(2),tmpB,1);
            % update VA and VB
    [~,Hyper.VA,~,Hyper.VB] = prior_MAR_NCP(p,kappa(1),kappa(2),Y0,Y);
    Hyper.iVA = sparse(1:n*p,1:n*p,1./Hyper.VA);
    Hyper.iVB = sparse(1:k*p,1:k*p,1./Hyper.VB);

    if strcmp(model,'BMAR-CSV') || strcmp(model,'BMAR-t') || strcmp(model,'BMAR-O')             
        Ec = YA - A'*XA*kron(speye(T),B);
        Er = Zc2Zr(Ec,k);
        iSigrEc = iSigr*Ec;
        iSigcEr = iSigc*Er;
        iSigcEr_c = Zc2Zr(iSigcEr,n);
        s2 = sum(reshape(sum(iSigcEr_c.*iSigrEc),k,T))'; % trace of iSigc*E_t*iSigr*E_t

        if strcmp(model,'BMAR-CSV') 
            % sample h
            if isim < 10
                h = sample_CSV(s2,phi,sigh2,h,n*k,true);
            else
                [h,is_accept] = sample_CSV(s2,phi,sigh2,h,n*k);
            end
            h = h - mean(h);
            count_h = count_h + is_accept;
            omega = exp(h);

            % sample phi and sigh2
            [~,phi,sigh2] = sample_SVpara(h,0,phi,Hyper);
        
        elseif strcmp(model,'BMAR-t')
            omega = 1./gamrnd((n*k+nu)/2,2./(s2+nu));

                % sample nu
            [nu,is_accept] = sample_nu(omega,nu,Hyper.nuub);
            count_nu = count_nu + is_accept;
        elseif strcmp(model,'BMAR-O')
                % sample o
            o_lpri = log([1-po; repmat(po/ngrid,ngrid,1)]);            
            for tt=1:T
                lliket = -n*k*log(o_grid) -.5*s2(tt)./o_grid.^2;
                o_post = exp(lliket + o_lpri - max(lliket));
                o_post = o_post/sum(o_post);
                idx = find(rand<cumsum(o_post),1);
                o(tt) = o_grid(idx);
            end
                % sample po
            tmp = sum(o>1);
            po = betarnd(Hyper.p0a + tmp, Hyper.p0b + T-tmp);
            omega = o.^2;
        end
    end
    iOmega = sparse(1:T,1:T,1./omega); 
    
    if isim > burnin        
        isave = isim-burnin;
        store_A(isave,:,:) = A;
        store_B(isave,:,:) = B;
        store_Sigc(isave,:,:) = Sigc;
        store_Sigr(isave,:,:) = Sigr;        
        store_omega(isave,:,:) = omega;
        store_kappa(isave,:) = kappa';
        if strcmp(model,'BMAR-CSV') 
            store_theta(isave,:) = [phi,sigh2];
        elseif strcmp(model,'BMAR-t') 
            store_theta(isave,:) = nu;
        elseif strcmp(model,'BMAR-O') 
            store_theta(isave,:) = po;
        end
    end
    
    if (mod(isim, 1000) ==0)
        disp([num2str(isim) ' loops... ' ])
    end     
end
disp( ['Estimation of ' model ' takes ' num2str(etime(clock, start_time)) ' seconds']);
disp(' ' );

A_mean = squeeze(mean(store_A));
B_mean = squeeze(mean(store_B));
Sigc_mean = squeeze(mean(store_Sigc));
Sigr_mean = squeeze(mean(store_Sigr));
omega_mean = squeeze(mean(store_omega))';
theta_mean = mean(store_theta)';
kappa_mean = mean(store_kappa)';

plotgraphs_MAR;

if cp_ml &&  strcmp(model,'BMAR') 
    [lml,lmlstd] = ml_BMAR(store_kappa,store_A,store_Sigr,Hyper,Y0,Y,XB,YB,R);
    fprintf('log marginal likelihood of BMAR: %.1f \n', lml);
end
