% This is the main run file for replicating the application in Chan and Qi (2025).
% This code estimates a BMAR with a factor SV 
%
% This code is free to use for academic purposes only, provided that the 
% paper is cited as:
%
% Chan, J.C.C. and Qi, Y. (2025). Large Bayesian Matrix Autoregressions, 
% Journal of Econometrics, forthcoming
%
% This code comes without technical support of any kind.  It is expected to
% reproduce the results reported in the paper. Under no circumstances will
% the authors be held responsible for any use (or misuse) of this code in
% any way.

clear; clc;
addpath('./utility') 
nsim = 10000; burnin = 500;
T0 = 8; 
p = 2; % # of lags
r = 6; % # of factors
dataset = 2; 
model = 'BMAR-FSV';
if dataset == 2 
    filename = 'Data_state_level_6q.xlsx';
    [data,regionID] = CleanData6q(filename);        
    MuY = mean(data,3);
    data = data - repmat(MuY,[1,1,size(data,3)]);
end
Y0 = data(:,:,1:T0);  % use the first T0 obs as initial values
Y = data(:,:,T0+1:end); 
[n,k,T] = size(Y);

X = zeros(n*p,k*p,T); 
for t = (T0+1):(T+T0)
    Xtcell = num2cell(data(:,:,t-1:-1:t-p),[1,2]);
    Xt = blkdiag(Xtcell{:});
    X(:,:,t-T0) = Xt;
end

    % priors
kappa = [.02,1]'; % kappaA, kappaB;
[Hyper.a0,Hyper.Va,Hyper.b0,Hyper.Vb] = prior_MAR_indep(p,kappa(1),kappa(2),Y0,Y);
Ca = Hyper.Va/kappa(1); % the constant part of Va
Cb = Hyper.Vb/kappa(2);
Hyper.iVa = sparse(1:n^2*p,1:n^2*p,1./Hyper.Va);
Hyper.iVb = sparse(1:k^2*p,1:k^2*p,1./Hyper.Vb);
Hyper.cA0 = [5,5]; 
Hyper.cB0 = [5,5]; 
Hyper.l0 = 0; Hyper.Vl = 1; % prior mean and variance for free elements in L
Hyper.nuh = 3*ones(n*k+r,1); Hyper.Sh = .1*(Hyper.nuh-1);
Hyper.mu0 = zeros(n*k+r,1); Hyper.Vmu = 100*ones(n*k+r,1);
Hyper.phi0 = .98*ones(n*k+r,1); Hyper.Vphi = .05^2*ones(n*k+r,1);

    % storage initialization
store_A = zeros(nsim,n*p,n);
store_B = zeros(nsim,k*p,k);
store_kappa = zeros(nsim,2);
store_l = zeros(nsim,n*k*r-r*(r+1)/2);
store_h = zeros(nsim,T,n*k+r);
store_hpara = zeros(nsim,3*(n*k+r)); % [mu' phi sig2];

start_time = clock;
     % MCMC starts here
disp(['Starting MCMC for ' model '...']);
randn('seed',sum(clock*100)); rand('seed',sum(clock*1000));    
    % initialize the Markov chain
F = randn(T,r);
L = [eye(r); zeros(n*k-r,r)];
B = repmat(eye(k),p,1);
A = zeros(n*p,n);
var_vecY = var(reshape(permute(Y,[3 1 2]),T,n*k))';
mu = [log(var_vecY/2); log(mean(var_vecY))*ones(r,1)];
phi = .9 + .09*rand(n*k+r,1);
sig2 = .05 + .05*rand(n*k+r,1);
h = [zeros(T,n*k) repmat(mu(n*k+1:end)',T,1)];

% construct a few things
    % for imposing the constraints M*[first column of B] = b0
M = sparse(1:p,1:k:k*p,ones(p,1),p,k*p);
b0 = ones(p,1);
count_phi = zeros(n*k+r,1);
L_idx = find(tril(ones(n*k,r),-1)~=0); % index of free elements of L

for isim = 1:nsim + burnin
        % sample A    
    Ebar = reshape(L*F',n,k,T);
    Ybar = Y - Ebar;
    ZBp = reshape(pagemtimes(X,repmat(B,1,1,T)),n*p,T*k); % [X_1 B, X_2 B, ...]
    for i=1:n
        iVai = Hyper.iVa((i-1)*n*p+1:i*n*p,(i-1)*n*p+1:i*n*p);
        ai0 = Hyper.a0((i-1)*n*p+1:i*n*p);
        Hi = h(:,i:n:n*k)';
        ZBpiDi = ZBp./repmat(exp(Hi(:)'),n*p,1);
        Kai = iVai + ZBpiDi*ZBp';
        CKai = chol(Kai,'lower');
        aihat = CKai'\(CKai\(iVai*ai0 + ZBpiDi*reshape(Ybar(i,:,:),T*k,1)));        
        ai = aihat + CKai'\randn(n*p,1);
        A(:,i) = ai;
    end  

    % sample B
    ZAp = reshape(pagemtimes(X,'transpose',repmat(A,1,1,T),'none'),k*p,T*n); % [X_1'A, X_2'A, ...]'
    for j=1:k
        iVbj = Hyper.iVb((j-1)*k*p+1:j*k*p,(j-1)*k*p+1:j*k*p);
        bj0 = Hyper.b0((j-1)*k*p+1:j*k*p);
        Hj = h(:,(j-1)*n+1:j*n)';        
        ZApiDj = ZAp./repmat(exp(Hj(:)'),k*p,1);
        Kbj = iVbj + ZApiDj*ZAp';
        CKbj = chol(Kbj,'lower');
        bjhat = CKbj'\(CKbj\(iVbj*bj0+ZApiDj*reshape(Ybar(:,j,:),T*n,1)));
        bj = bjhat + CKbj'\randn(k*p,1);
        if j==1
            tmpU = CKbj'\(CKbj\M');
            tmpV = (M*tmpU)\tmpU';
            bj = bj - tmpV'*(M*bj-b0);
        end
        B(:,j) = bj; 
    end    

        % sample kappaA and kappaB
    tmpa = ((A(:)-Hyper.a0)./Ca)'*(A(:)-Hyper.a0);
    kappa(1) = gigrnd(Hyper.cA0(1)-n^2*p/2,2*Hyper.cA0(2),tmpa,1);
    tmpb = ((B(:)-Hyper.b0)./Cb)'*(B(:)-Hyper.b0);    
    kappa(2) = gigrnd(Hyper.cB0(1)-k^2*p/2,2*Hyper.cB0(2),tmpb,1);
        % update Va and Vb
    [~,Hyper.Va,~,Hyper.Vb] = prior_MAR_indep(p,kappa(1),kappa(2),Y0,Y);
    Hyper.iVa = sparse(1:n^2*p,1:n^2*p,1./Hyper.Va);
    Hyper.iVb = sparse(1:k^2*p,1:k^2*p,1./Hyper.Vb);

        % sample F
    E = Y - pagemtimes(A,'transpose',pagemtimes(X,repmat(B,1,1,T)),'none');
    e = reshape(E,n*k*T,1);    
    Xf = kron(speye(T),L);
    XfiSig = Xf'*sparse(1:T*n*k,1:T*n*k,reshape(exp(-h(:,1:n*k))',T*n*k,1)); 
    Kf = sparse(1:T*r,1:T*r,reshape(exp(-h(:,n*k+1:end))',T*r,1)) + XfiSig*Xf;
    f_hat = Kf\(XfiSig*e);
    f = f_hat + chol(Kf,'lower')'\randn(T*r,1);
    F = reshape(f,r,T)';

        % sample L
    for ii = 2:n*k
        if ii <= r
            Xli = F(:,1:ii-1);
            iVli = sparse(1:ii-1,1:ii-1,1/Hyper.Vl*ones(ii-1,1)); 
            XliOmega = Xli'*sparse(1:T,1:T,exp(-h(:,ii)));
            dli = iVli*Hyper.l0*ones(ii-1,1) + XliOmega*(e(ii:n*k:end)-F(:,ii)); 
        else
            Xli = F;
            iVli = sparse(1:r,1:r,1/Hyper.Vl*ones(r,1));            
            XliOmega = Xli'*sparse(1:T,1:T,exp(-h(:,ii)));
            dli = iVli*Hyper.l0*ones(r,1) + XliOmega*e(ii:n*k:end);
        end
        Kli = iVli + XliOmega*Xli;
        CKli = chol(Kli,'lower');
        li_hat = CKli'\(CKli\dli);            
        li = li_hat + CKli'\randn(min(ii-1,r),1);       
        L(ii,1:min(ii-1,r)) = li;
    end

        % sample h 
    Uy = E - reshape(L*F',n,k,T);
    vecUy = reshape(permute(Uy,[3 1 2]),T,n*k);
    ystar = log([vecUy F].^2 + 1e-4);
    for jj=1:n*k+r
        h(:,jj) = sample_SV(ystar(:,jj),h(:,jj),mu(jj),phi(jj),sig2(jj));
    end 

        % sample mu, phi and sig2
    [mu,phi,sig2,flag_phi] = sample_SVpara(h,mu,phi,Hyper);
    count_phi = count_phi + flag_phi; 

    
    if isim > burnin        
        isave = isim-burnin;
        store_A(isave,:,:) = A;
        store_B(isave,:,:) = B;
        store_kappa(isave,:) = kappa';
        store_l(isave,:) = L(L_idx);
        store_h(isave,:,:) = h;        
        store_hpara(isave,:) =  [mu',phi',sig2']; 
    end
    
    if (mod(isim, 1000) ==0)
        disp([num2str(isim) ' loops... ' ])
    end     
end
disp( ['Estimation of ' model ' takes ' num2str(etime(clock, start_time)) ' seconds']);
disp(' ' );

A_mean = squeeze(mean(store_A));
B_mean = squeeze(mean(store_B));

hpara_mean = mean(store_hpara)';
kappa_mean = mean(store_kappa)';
l_mean = mean(store_l)';
h_mean = squeeze(mean(store_h));


plotgraphs_MAR;


