% This function computes the marginal likelihood of BMAR
%
% See:
% Chan, J.C.C. and Qi, Y. (2025). Large Bayesian Matrix Autoregressions, 
% Journal of Econometrics, forthcoming

function [lml,lmlstd] = ml_BMAR(store_kappa,store_A,store_Sigr,Hyper,Y0,Y,XB,YB,R)
R = 50*ceil(R/50);
nsim = size(store_kappa,1);
[n,k,T] = size(Y);
p = size(store_A,2)/n;

big_kappa = zeros(R,2);
ckappa_hat = zeros(2,2);
for i=1:2
    tmp_hat = gamfit(store_kappa(:,i));
    ckappa_hat(i,:) = [tmp_hat(1) 1/tmp_hat(2)];
    big_kappa(:,i) = max(gamrnd(ckappa_hat(i,1),1./ckappa_hat(i,2),R,1),1e-10);
end

storeA1 = reshape(store_A,nsim,n^2*p);
ahat = mean(storeA1)'; avar = cov(storeA1); Cavar = chol(avar,'lower');
big_a = repmat(ahat',R,1) + (Cavar*randn(n^2*p,R))';

tmpsum = zeros(n,n);
for isim = 1:nsim
    Sigr = squeeze(store_Sigr(isim,:,:));
    tmpsum = tmpsum + Sigr\speye(n);
end
nurhat = T*k;
Srhat = (tmpsum/(nurhat*nsim))\speye(n);

store_w = zeros(R,1);
lpri_kappa  = @(xk) lgampdf(xk(1),Hyper.cA0(1),Hyper.cA0(2)) ...
    + lgampdf(xk(2),Hyper.cB0(1),Hyper.cB0(2));
gIS = @(xa,xS,xk) lmvnpdf(xa,ahat,avar) + linvwishpdf(xS,nurhat,Srhat) ...
    +lgampdf(xk(1),ckappa_hat(1,1),ckappa_hat(1,2)) ...
    +lgampdf(xk(2),ckappa_hat(2,1),ckappa_hat(2,2));

for isim = 1:R
    Sigr = iwishrnd(Srhat,nurhat);
    iSigr = Sigr\speye(n);
    kappa = big_kappa(isim,:);
    a = big_a(isim,:)';
    A = reshape(a,n*p,n);
    
    [~,Hyper.VA,~,Hyper.VB] = prior_MAR_NCP(p,kappa(1),kappa(2),Y0,Y);
    Hyper.iVA = sparse(1:n*p,1:n*p,1./Hyper.VA); 
    Hyper.iVB = sparse(1:k*p,1:k*p,1./Hyper.VB);
    KB = Hyper.iVB + XB*kron(speye(T),A*iSigr*A')*XB';
    CKB = chol(KB,'lower');
    Bhat = CKB'\(CKB\(Hyper.iVB*Hyper.B0 + XB*kron(speye(T),A*iSigr)*YB'));
    Schat = Hyper.Sc + Hyper.B0'*Hyper.iVB*Hyper.B0 +YB*kron(speye(T),iSigr)*YB' -Bhat'*KB*Bhat; 
    Schat = (Schat+Schat')/2;
    llike = intlike_BMAR(Sigr,KB,Schat,Hyper,T);

    w = llike +lniwpdf(A,Sigr,Hyper.A0,Hyper.iVA,Hyper.nur,Hyper.Sr) +lpri_kappa(kappa)...
        -gIS(a,Sigr,kappa);
    store_w(isim) = w;    
end
shortw = reshape(store_w,R/50,50);
maxw = max(shortw);

bigml = log(mean(exp(shortw-repmat(maxw,R/50,1)),1)) + maxw;
lml = mean(bigml);
lmlstd = std(bigml)/sqrt(50); 
 
% 
% [~,Hyper.VA,~,Hyper.VB] = prior_MAR_NCP(p,kappa_mean(1),kappa_mean(2),Y0,Y);
% Hyper.iVA = sparse(1:n*p,1:n*p,1./Hyper.VA); Hyper.iVB = sparse(1:k*p,1:k*p,1./Hyper.VB);
% KB = Hyper.iVB + XB*kron(speye(T),A_mean*iSigr_mean*A_mean')*XB';
% CKB = chol(KB,'lower');
% Bhat = CKB'\(CKB\(Hyper.iVB*Hyper.B0 + XB*kron(speye(T),A_mean*iSigr_mean)*YB'));
% Schat = Hyper.Sc + Hyper.B0'*Hyper.iVB*Hyper.B0 + YB*kron(speye(T),iSigr_mean)*YB' - Bhat'*KB*Bhat; 
% Schat = (Schat+Schat')/2;
% llike = intlike_BMAR(Sigr_mean,KB,Schat,Hyper,T);
% 
% lprior = lniwpdf(A_mean,Sigr_mean,Hyper.A0,Hyper.iVA,Hyper.nur,Hyper.Sr);
% 
%     % evaluate the posterior density
% store_lpost = zeros(nsim,1); % [log density of (A,Sigr)
% 
% for isim = 1:nsim
%     B = squeeze(store_B(isim,:,:));
%     Sigc = squeeze(store_Sigc(isim,:,:));
%     iSigc = Sigc\speye(k);
% 
%         % compute the conditional density of Sigr and A 
%     KA = Hyper.iVA + XA*kron(speye(T),B*iSigc*B')*XA';
%     CKA = chol(KA,'lower');
%     Ahat = CKA'\(CKA\(Hyper.iVA*Hyper.A0 + XA*kron(speye(T),B*iSigc)*YA'));    
%     Srhat = Hyper.Sr + Hyper.A0'*Hyper.iVA*Hyper.A0 + YA*kron(iOmega,iSigc)*YA' - Ahat'*KA*Ahat;    
%     Srhat = (Srhat+Srhat')/2;
%     lden_ASigr = lniwpdf(A_mean,Sigr_mean,Ahat,KA,Hyper.nur+T*k,Srhat);
% 
%     store_lpost(isim,:) = lden_ASigr;
% end
% tmpmax = max(store_lpost);
% lpost = log(mean(exp(store_lpost-repmat(tmpmax,nsim,1)))) + tmpmax;


% QA = diag((A_mean-Hyper.A0)*(Sigr_mean\(A_mean-Hyper.A0)'));
% tmpA = sum(QA./CA);
% QB = diag((B_mean-Hyper.B0)*(Sigc_mean\(B_mean-Hyper.B0)'));
% tmpB = sum(QB./CB);
% lgigpdf(kappa_mean(1),Hyper.cA0(1)-n^2*p/2,2*Hyper.cA0(2),tmpA)
% lgigpdf(kappa_mean(2),Hyper.cB0(1)-k^2*p/2,2*Hyper.cB0(2),tmpB)
% 
% 
% lgigpdf(x,a,b,p)

% lml = llike + lprior - lpost;
fprintf('log marginal likelihood of BMAR: %.1f \n', lml);
end